/*
 * Copyright (C) 2021 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 *
 * Description:
 * Performance measurements of PowerFIP library
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>     // memset()
#include <stdint.h>     // uint8_t, ...
#include <fcntl.h>      // open()
#include <unistd.h>     // read()
#include <libgen.h>     // basename
#include <inttypes.h>   // PRIx64+
#include <signal.h>     // signal
#include <time.h>       // nanosleep, timespec
#if defined (_WIN32) || defined (_WIN64)
#include <conio.h>      // kbhit()
#endif

#include "common.h"

char *app_name = NULL;
int running = 0;

static void help();
void handle_signal(int sig);

int main(int argc, char *argv[])
{
    int ret = 0;
    int dev_idx = 1;
    int station_opt = 0;
    int bitrate_opt = 1;
    int value = 0;
    struct timespec rqtp = {0};
    struct timespec rmtp = {0};
    struct tst_session sess;
    char *erase_end_line;

    app_name = basename(argv[0]);

#if defined (_WIN32) || defined (_WIN64)
    SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
    enable_console_colorization();
#else
    uint8_t buff = 0;
#endif

    /* Try to process all command line arguments */
    while ((value = getopt(argc, argv, "i:s:b:lhv")) != -1) {
        switch (value) {
            case 'i':
                sscanf(optarg, "%d", &dev_idx);
                break;
            case 's':
                sscanf(optarg, "%d", &station_opt);
                if ((station_opt < 0) || (station_opt > 1)) {
                    help();
                    return EXIT_FAILURE;
                }
                break;
            case 'b':
                sscanf(optarg, "%d", &bitrate_opt);
                if (bitrate_opt > 3) {
                    help();
                    return EXIT_FAILURE;
                }
                break;
            case 'l':
                dev_show_all();
                return EXIT_SUCCESS;
            case 'h':
                help();
                return EXIT_SUCCESS;
            case 'v':
                printf("%s %s\n", app_name, APP_VERSION);
                return EXIT_SUCCESS;
            case '?':
            default:
                help();
                return EXIT_FAILURE;
        }
    }

    signal(SIGINT, handle_signal);

    /* pwrfip - initialization [mandatory] */
    if (pwrfip_init()) {
        _print(app, err, "cannot init pwrfip library: %s\n",
            pwrfip_strerror(errno));
        return -1;
    }

    /* exoligent - fip device opening */
    memset(&sess, 0, sizeof(struct tst_session));
    if ((ret = dev_open(dev_idx, &sess)))
        goto fail1;

    sess.station_number = station_opt;
    sess.station_bitrate = bitrate_opt;
    if (tst_session_start(&sess) < 0) {
        _print(app, err, "impossible to start fip station %d.\n", station_opt);
        goto fail2;
    }

    _print(app, info, "app started. press enter to close...\n");
    erase_end_line = (dbg_coloriz) ? "\033[0K" : "";

#if defined(__linux__) || defined(__QNX__)
    /* set stdin non-blocking */
    fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
#endif

    rqtp.tv_sec = 0;
    rqtp.tv_nsec = 500000000; /* 500ms */

    /* main loop */
    running = 1;
    while (running) {
        /* wait for a pressed key */
#if defined (_WIN32) || defined (_WIN64)
        if (kbhit())
#else
        if (read(STDIN_FILENO, &buff, 1) == 1)
#endif
        {
            running = 0;
        }

        _print(sess.station_number, info,
            "=> ba[%s]:mcycle=%"PRIu64"%s\r",
            pwrfip_ba_state_str[sess.node.ba_status.state],
            sess.stats.nb_sync_cb, erase_end_line);
        fflush(stdout);

        ret = nanosleep(&rqtp, &rmtp);
    }

    if (tst_session_stop(&sess))
        _print(app, err, "impossible to stop fip station %d.\n", station_opt);

    {
        FILE* report_file = NULL;
        char report_path[255];

        _print(app, info, "report is building ...\n");
        memset(report_path, 0, sizeof(report_path));
        sprintf(report_path, "./report_idx%d.txt", dev_idx);
        report_file = fopen(report_path, "wb");
        print_stats(report_file, &sess);
        if (report_file) {
            _print(app, info, "report is saved to: %s\n", report_path);
            fclose(report_file);
        }
    }

fail2:
    /* exoligent - fip device closure */
    if (dev_close(&sess))
        _print(app, err, "cannot close correctly the device.\n");
fail1:
    pwrfip_exit();
    _print(app, info, "app closed\n");
    return ret;
}

/**
 * \brief Callback function for handling signals.
 * \param sig identifier of signal
 */
void handle_signal(int sig)
{
    if (sig == SIGINT) {
        _print(app, info, "SIGINT signal received\n");
        running = 0;
        /* Reset signal handling to default behavior */
        signal(SIGINT, SIG_DFL);
    } else
        _print(app, warn, "signal received but not supported (sig=%d)\n", sig);
}

/**
 * \brief Print help for this application
 */
static void help()
{
    printf("\n");
    printf("Usage: %s [OPTION]...\n\n", app_name);
    printf("It tests FIP board performance with PowerFIP library.\n");
    printf("By default, if no option is added, the app opens the first PCI/PCIe device\n");
    printf("with index 1 [-i 1], and starts FIP node 0 (addr=0) [-s 0] with bitrate sets\n");
    printf("to 1Mbps [-b 1].\n\n");
    printf("Options:\n");
    printf("    -i   device index [default=1]\n");
    printf("    -s   FIP stations to start [default=0]\n");
    printf("         0: station0\n");
    printf("         1: station1\n");
    printf("    -b   FIP bitrate [default=1]\n");
    printf("         0: 31.25Kbps\n");
    printf("         1: 1Mbps\n");
    printf("         2: 2.5Mbps\n");
    printf("         3: 5Mbps\n");
    printf("    -l   list the FIP boards present on the host machine\n");
    printf("    -h   show this help and exit\n");
    printf("    -v,  show version and exit\n");
    printf("\n");
    printf("Examples: \n");
    printf("    %s -i 1 -s 0 -b 1\n", app_name);
    printf("\n");
}